@@ -948,8 +948,9 @@ This function is for internal use only."
 
 (defun epg-signature-to-string (signature)
   "Convert SIGNATURE to a human readable string."
-  (let ((user-id (cdr (assoc (epg-signature-key-id signature)
-			     epg-user-id-alist))))
+  (let* ((user-id (cdr (assoc (epg-signature-key-id signature)
+			      epg-user-id-alist)))
+	 (pubkey-algorithm (epg-signature-pubkey-algorithm signature)))
     (concat
      (cond ((eq (epg-signature-status signature) 'good)
 	    "Good signature from ")
@@ -972,6 +973,15 @@ This function is for internal use only."
        "")
      (if (epg-signature-validity signature)
 	 (format " (trust %s)"  (epg-signature-validity signature))
+       "")
+     (if (epg-signature-creation-time signature)
+	 (format-time-string " created at %Y-%m-%dT%T%z"
+			     (epg-signature-creation-time signature))
+       "")
+     (if pubkey-algorithm
+	 (concat " using "
+		 (or (cdr (assq pubkey-algorithm epg-pubkey-algorithm-alist))
+		     (format "(unknown algorithm %d)" pubkey-algorithm)))
        ""))))
 
 (defun epg-verify-result-to-string (verify-result)
@@ -1368,9 +1378,10 @@ This function is for internal use only."
       (epg-signature-set-creation-time
        signature
        (epg--time-from-seconds (match-string 2 string)))
-      (epg-signature-set-expiration-time
-       signature
-       (epg--time-from-seconds (match-string 3 string)))
+      (unless (equal (match-string 3 string) "0")
+	(epg-signature-set-expiration-time
+	 signature
+	 (epg--time-from-seconds (match-string 3 string))))
       (epg-signature-set-version
        signature
        (string-to-number (match-string 4 string)))
@@ -1661,7 +1672,8 @@ This function is for internal use only."
    (string-to-number (aref line 2))
    (aref line 4)
    (epg--time-from-seconds (aref line 5))
-   (epg--time-from-seconds (aref line 6))))
+   (if (aref line 6)
+       (epg--time-from-seconds (aref line 6)))))
 
 ;;;###autoload
 (defun epg-list-keys (context &optional name mode)
@@ -2080,20 +2092,37 @@ Otherwise, it makes a cleartext signature."
 If optional 3rd argument MODE is t or 'detached, it makes a detached signature.
 If it is nil or 'normal, it makes a normal signature.
 Otherwise, it makes a cleartext signature."
-  (unwind-protect
-      (progn
-	(epg-context-set-output-file context
-				     (epg--make-temp-file "epg-output"))
-	(epg-start-sign context (epg-make-data-from-string plain) mode)
-	(epg-wait-for-completion context)
-	(unless (epg-context-result-for context 'sign)
-	  (if (epg-context-result-for context 'error)
-	      (error "Sign failed: %S"
-		     (epg-context-result-for context 'error))
-	    (error "Sign failed")))
-	(epg-read-output context))
-    (epg-delete-output-file context)
-    (epg-reset context)))
+  (let ((input-file
+	 (unless (or (eq (epg-context-protocol context) 'CMS)
+		     (condition-case nil
+			 (progn
+			   (epg-check-configuration (epg-configuration))
+			   t)
+		       (error)))
+	   (epg--make-temp-file "epg-input")))
+	(coding-system-for-write 'binary))
+    (unwind-protect
+	(progn
+	  (epg-context-set-output-file context
+				       (epg--make-temp-file "epg-output"))
+	  (if input-file
+	      (write-region plain nil input-file nil 'quiet))
+	  (epg-start-sign context
+			  (if input-file
+			      (epg-make-data-from-file input-file)
+			    (epg-make-data-from-string plain))
+			  mode)
+	  (epg-wait-for-completion context)
+	  (unless (epg-context-result-for context 'sign)
+	    (if (epg-context-result-for context 'error)
+		(error "Sign failed: %S"
+		       (epg-context-result-for context 'error))
+	      (error "Sign failed")))
+	  (epg-read-output context))
+      (epg-delete-output-file context)
+      (if input-file
+	  (delete-file input-file))
+      (epg-reset context))))
 
 ;;;###autoload
 (defun epg-start-encrypt (context plain recipients
@@ -2181,25 +2210,42 @@ If RECIPIENTS is nil, it performs symmetric encryption."
 				   &optional sign always-trust)
   "Encrypt a string PLAIN.
 If RECIPIENTS is nil, it performs symmetric encryption."
-  (unwind-protect
-      (progn
-	(epg-context-set-output-file context
-				     (epg--make-temp-file "epg-output"))
-	(epg-start-encrypt context (epg-make-data-from-string plain)
-			   recipients sign always-trust)
-	(epg-wait-for-completion context)
-	(if (and sign
-		 (not (epg-context-result-for context 'sign)))
-	    (if (epg-context-result-for context 'error)
-		(error "Sign failed: %S"
-		       (epg-context-result-for context 'error))
-	      (error "Sign failed")))
-	(if (epg-context-result-for context 'error)
-	    (error "Encrypt failed: %S"
-		   (epg-context-result-for context 'error)))
-	(epg-read-output context))
-    (epg-delete-output-file context)
-    (epg-reset context)))
+  (let ((input-file
+	 (unless (or (not sign)
+		     (eq (epg-context-protocol context) 'CMS)
+		     (condition-case nil
+			 (progn
+			   (epg-check-configuration (epg-configuration))
+			   t)
+		       (error)))
+	   (epg--make-temp-file "epg-input")))
+	(coding-system-for-write 'binary))
+    (unwind-protect
+	(progn
+	  (epg-context-set-output-file context
+				       (epg--make-temp-file "epg-output"))
+	  (if input-file
+	      (write-region plain nil input-file nil 'quiet))
+	  (epg-start-encrypt context
+			     (if input-file
+				 (epg-make-data-from-file input-file)
+			       (epg-make-data-from-string plain))
+			     recipients sign always-trust)
+	  (epg-wait-for-completion context)
+	  (if (and sign
+		   (not (epg-context-result-for context 'sign)))
+	      (if (epg-context-result-for context 'error)
+		  (error "Sign failed: %S"
+			 (epg-context-result-for context 'error))
+		(error "Sign failed")))
+	  (if (epg-context-result-for context 'error)
+	      (error "Encrypt failed: %S"
+		     (epg-context-result-for context 'error)))
+	  (epg-read-output context))
+      (epg-delete-output-file context)
+      (if input-file
+	  (delete-file input-file))
+      (epg-reset context))))
 
 ;;;###autoload
 (defun epg-start-export-keys (context keys)
@@ -2437,7 +2483,7 @@ PARAMETERS is a string which tells how to create the key."
 (defun epg--decode-hexstring (string)
   (let ((index 0))
     (while (eq index (string-match "[0-9A-Fa-f][0-9A-Fa-f]" string index))
-      (setq string (replace-match "\\x\\&" t nil string)
+      (setq string (replace-match "\\\\x\\&" t nil string)
 	    index (+ index 4)))
     (car (read-from-string (concat "\"" string "\"")))))
 
@@ -2450,7 +2496,7 @@ PARAMETERS is a string which tells how to create the key."
 	  (setq string (replace-match "\\2" t nil string)
 		index (1+ index))
 	(if (match-beginning 3)
-	    (setq string (replace-match "\\x\\3" t nil string)
+	    (setq string (replace-match "\\\\x\\3" t nil string)
 		  index (+ index 4))
 	  (setq string (replace-match "\\\\\\\\\\4" t nil string)
 		index (+ index 3)))))
